// 
// ------------------------------------------------------------------------
// Copyright (C) 2012, 
// University of Perpignan Via Domitia, DALI.
//
//   This file is part of Perpi.
//
//   Perpi is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   Perpi is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with Perpi.  If not, see <http://www.gnu.org/licenses/>.
//
//   Authors: 
//       Bernard Goossens (goossens@univ-perp.fr)
//       David Parello (david.parello@univ-perp.fr)
// ------------------------------------------------------------------------
//

#ifndef _COMMON_CORES_MODULES_ISSUEUNIT_TCC_
#define _COMMON_CORES_MODULES_ISSUEUNIT_TCC_
//#define DEBUGBG
#include <pin.H>
#include <instruction.hh>
#include <common/cores/modules/issueunit.hh>

namespace common{
  namespace cores{
    namespace modules{

template <class CONFIG>
IssueUnit<CONFIG>::IssueUnit() 
{
  first_free_entry = 0;
  first_occupied_entry = 0;
  empty = true;
}

template <class CONFIG>
uint64_t IssueUnit<CONFIG>::issue(Instruction *ins, uint64_t cycle) 
{
  return (cycle+CONFIG::ISSUE::Latency);
}

template <class CONFIG>
void IssueUnit<CONFIG>::remove_min_full_issue_cycles(uint64_t cycle) {
  uint32_t i;
  for (i=first_occupied_entry; i!=first_free_entry &&
         ((nb_issued_in_same_cycle[i] == CONFIG::ISSUE::IssueWidth)
      || (issue_cycle[i]<cycle));){
    i++;
    if (i == PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES) i=0;
  }
/*
  for (j=i; j<first_free_entry; j++){
    nb_issued_in_same_cycle[j-i] = nb_issued_in_same_cycle[j];
    issue_cycle[j-i] = issue_cycle[j];
  }
  first_free_entry = first_free_entry - i;
*/
  first_occupied_entry = i;
  if (first_free_entry == first_occupied_entry){
    first_free_entry = 0;
    first_occupied_entry = 0;
    empty = true;
  }
}

template <class CONFIG>
void IssueUnit<CONFIG>::insert_max_issue_cycle(uint64_t cycle) {
  uint64_t c, pre, nb;
  if (empty){
    issue_cycle[0] = cycle;
    nb_issued_in_same_cycle[0] = 1;
    first_free_entry=1;
    empty = false;
    return;
  }
  if (first_free_entry == 0)
    pre = PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES - 1;
  else
    pre = first_free_entry - 1;
  c = issue_cycle[pre]+1;
  if (first_free_entry>=first_occupied_entry)
    nb = first_free_entry - first_occupied_entry;
  else
    nb = (first_free_entry + PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES) - first_occupied_entry;
  if (nb >= PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES){
        cerr << "too many issue cycles while inserting " << cycle << " cycle at end" << endl;
        cerr << "first_free_entry: " << first_free_entry << endl;
        abort();
  }
  for (; c<cycle;){
    issue_cycle[first_free_entry] = c;
    nb_issued_in_same_cycle[first_free_entry] = 0;
    first_free_entry++;
    if (first_free_entry == PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES)
      first_free_entry = 0;
    c++;
  }
  issue_cycle[first_free_entry] = cycle;
  nb_issued_in_same_cycle[first_free_entry] = 1;
  first_free_entry++;
  if (first_free_entry == PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES)
    first_free_entry = 0;
  empty = false;
}

template <class CONFIG>
void IssueUnit<CONFIG>::dump(uint64_t cycle){
  cerr << "first_occupied_entry: " << first_occupied_entry << endl;
  cerr << "first_free_entry: " << first_free_entry << endl;
  cerr << "issue cycle: " << issue_cycle[cycle] << endl;
  for (uint32_t i=first_occupied_entry; i!=first_free_entry; ){
    cerr << "cycle: " << issue_cycle[i];
    cerr << " nb_issues: " << nb_issued_in_same_cycle[i] << endl;
    i++;
    if (i == PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES) i = 0;
  }
}

template <class CONFIG>
uint64_t IssueUnit<CONFIG>::get_issue_cycle(uint64_t previous_fetch_cycle, uint64_t cycle) {
  uint64_t i, p, c, nb, difference;
  remove_min_full_issue_cycles(previous_fetch_cycle);
  if (empty){
    insert_max_issue_cycle(cycle);
#ifdef DEBUGBG
    dump(0);
#endif
    return cycle;
  }
  p = look_for_issue_cycle_position(cycle);
  //not found and the searched cycle is larger than the largest: add all the intermediate cycles at end
  if (p == first_free_entry){
    insert_max_issue_cycle(cycle);
#ifdef DEBUGBG
    dump(first_free_entry-1);
#endif
    return cycle;
  }
  //found and not full cycle: increment
  if (issue_cycle[p] == cycle && nb_issued_in_same_cycle[p]<CONFIG::ISSUE::IssueWidth){
    nb_issued_in_same_cycle[p]++;
#ifdef DEBUGBG
    dump(p);
#endif
    return cycle;
  }
  if (first_free_entry>=first_occupied_entry)
    nb = first_free_entry - first_occupied_entry;
  else
    nb = first_free_entry + PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES - first_occupied_entry;
  //found and full cycle: look for the next non full cycle
  if (issue_cycle[p] == cycle){
    i = p+1;
    if (i >= PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES)
        i = i - PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES;
    for (; i!=first_free_entry && (nb_issued_in_same_cycle[i]==CONFIG::ISSUE::IssueWidth);){
      i++;
      if (i == PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES) i = 0;
    }
    //open a new issue cycle
    if (i == first_free_entry){
      //no place free
      if (nb >= PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES){
        cerr << "too many issue cycles while inserting " << cycle << " at position " << p << endl;
        abort();
      }
      //new entry, initialize cycle number, one instruction issued
      issue_cycle[i] = cycle+i-p;
      nb_issued_in_same_cycle[i] = 1;
      first_free_entry++;
      if (first_free_entry == PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES)
        first_free_entry = 0;
      empty = false;
    }
    //one more issue at this cycle
    else nb_issued_in_same_cycle[i]++;
#ifdef DEBUGBG
    dump(i);
#endif
    //return the issue cycle found
    return issue_cycle[i];
  }
  if (issue_cycle[p] < cycle){
    cerr << "strange cycle: " << cycle << endl;
    cerr << "to be inserted at position: " << p << endl;
    abort();
  }
  if (p != first_occupied_entry){
    cerr << "strange position: " << p << endl;
    abort();
  }
  //not found and the searched cycle is smaller than the smallest: add all the intermediate cycles at start
  c = issue_cycle[first_occupied_entry];
  difference = c - cycle;
  if ((nb+difference) >= PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES){
    cerr << "too many issue cycles while inserting " << cycle << " cycle at start" << endl;
    cerr << "first_occupied_entry: " << first_occupied_entry << endl;
    cerr << "first_free_entry: " << first_free_entry << endl;
    cerr << "nb entries: " << nb << endl;
    cerr << "difference: " << difference << endl;
    cerr << "cycle: " << cycle << endl;
    cerr << "c: " << c << endl;
    cerr << "p: " << p << endl;
    abort();
  }
/*
  //move right to create c-cycle free entries
  for (; difference>0; difference--){
    issue_cycle[c+difference-1] = issue_cycle[c-1];
    nb_issued_in_same_cycle[c+difference-1] = nb_issued_in_same_cycle[c-1];
  }
  nb_issued_in_same_cycle[0] = 1;
  issue_cycle[0] = cycle;
*/
  for (; c>cycle; c--){
    if (first_occupied_entry == 0)
      first_occupied_entry = PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES-1;
    else
      first_occupied_entry = first_occupied_entry-1;
    issue_cycle[first_occupied_entry] = c-1;
    nb_issued_in_same_cycle[first_occupied_entry] = 0;
  }
  if (first_occupied_entry == 0)
    first_occupied_entry = PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES-1;
  else
    first_occupied_entry = first_occupied_entry-1;
  nb_issued_in_same_cycle[first_occupied_entry] = 1;
  issue_cycle[first_occupied_entry] = cycle;
  empty = false;
  return cycle;
}

template <class CONFIG>
uint64_t IssueUnit<CONFIG>::look_for_issue_cycle_position(uint64_t cycle){
  uint64_t d, m, mm, f;
  if (empty) return 0;
  d=first_occupied_entry;
  if (first_free_entry<=first_occupied_entry)
    f=first_free_entry+PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES;
  else 
    f=first_free_entry;
  while (d!=f-1){
    m=d+(f-d)/2;
    if (m>=PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES)
      mm=m-PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES;
    else
      mm=m;
    if (cycle<issue_cycle[mm]) f=m;
    else d=m;
  }
  if (d>=PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES)
    d = d-PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES;
  if (cycle<=issue_cycle[d]) return d;
  if (f>=PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES)
    f = f-PERPICONFIG::INSTRUCTION::NB_MAX_ISSUE_CYCLES;
  return f;
}

#undef DEBUGBG

    }//    namespace modules{
  }//  namespace cores{
}// namespace common{


#endif // _COMMON_CORES_MODULES_ISSUEUNIT_TCC_
